@@ -1,4 +1,4 @@
-# Copyright (C) 2015-2022 S[&]T, The Netherlands.
+# Copyright (C) 2015-2023 S[&]T, The Netherlands.
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -30,6 +30,7 @@
 from __future__ import print_function
 
 from collections import OrderedDict
+import csv
 import datetime
 import glob
 import numpy
@@ -851,7 +852,7 @@ def _import_scalar(c_data_type, c_data):
 
 def _import_array(c_data_type, c_num_elements, c_data):
     if c_data_type == _lib.harp_type_string:
-        data = numpy.empty((c_num_elements,), dtype=numpy.object)
+        data = numpy.empty((c_num_elements,), dtype=numpy.object_)
         for i in range(c_num_elements):
             # NB. The _ffi.string() method returns a copy of the C string.
             data[i] = _decode_string(_ffi.string(c_data.string_data[i]))
@@ -1192,8 +1193,8 @@ def import_product(filename, operations="", options="", reduce_operations="", po
     be appended into a single merged product and that merged product will be returned.
 
     Arguments:
-    filename -- Filename, file pattern, or list of filenames/patterns of the product(s)
-                to import
+    filename -- Filename, file pattern, .pth file, or list of filenames/patterns/.pths of
+                the product(s) to import
     operations -- Actions to apply as part of the import; should be specified as a
                   semi-colon separated string of operations;
                   in case a list of products is ingested these operations will be
@@ -1226,8 +1227,22 @@ def import_product(filename, operations="", options="", reduce_operations="", po
     elif '*' in filename or '?' in filename:
         # This is a globbing pattern
         filenames = sorted(glob.glob(filename))
+    elif filename.endswith('.pth'):
+        filenames = [filename]
 
     if filenames is not None:
+        expanded_filenames = []
+        for filename in filenames:
+            if filename.endswith('.pth'):
+                with open(filename, newline='') as csvfile:
+                    pathreader = csv.reader(csvfile)
+                    next(pathreader)  # skip header
+                    for row in pathreader:
+                        expanded_filenames.append(row[0])
+            else:
+                expanded_filenames.append(filename)
+        filenames = expanded_filenames
+
         if len(filenames) == 0:
             raise Error("no files matching '%s'" % (filename))
         # Return the merged concatenation of all products
